# encoding: utf-8

"""Unit test suite for `cr.cube.stripe.assembler` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _Element, _OrderSpec, _Subtotal
from cr.cube.enums import COLLATION_METHOD as CM
from cr.cube.stripe.assembler import (
    StripeAssembler,
    _BaseOrderHelper,
    _BaseSortByValueHelper,
    _OrderHelper,
    _SortByLabelHelper,
    _SortByMeasureHelper,
)
from cr.cube.stripe.measure import (
    StripeMeasures,
    _BaseSecondOrderMeasure,
    _Means,
    _PopulationProportions,
    _PopulationProportionStderrs,
    _ScaledCounts,
    _TableProportions,
    _TableProportionStddevs,
    _TableProportionStderrs,
    _UnweightedBases,
    _UnweightedCounts,
    _WeightedBases,
    _WeightedCounts,
)

from ...unitutil import class_mock, instance_mock, method_mock, property_mock


class DescribeStripeAssembler:
    """Unit test suite for `cr.cube.stripe.assembler.StripeAssembler` object."""

    @pytest.mark.parametrize(
        "measure_prop_name, MeasureCls",
        (
            ("means", _Means),
            ("population_proportions", _PopulationProportions),
            ("population_proportion_stderrs", _PopulationProportionStderrs),
            ("table_proportion_stddevs", _TableProportionStddevs),
            ("table_proportion_stderrs", _TableProportionStderrs),
            ("table_proportions", _TableProportions),
            ("unweighted_bases", _UnweightedBases),
            ("unweighted_counts", _UnweightedCounts),
            ("weighted_bases", _WeightedBases),
            ("weighted_counts", _WeightedCounts),
        ),
    )
    def it_assembles_various_measures(
        self,
        request,
        _measures_prop_,
        measures_,
        _assemble_vector_,
        measure_prop_name,
        MeasureCls,
    ):
        _measures_prop_.return_value = measures_
        setattr(
            measures_,
            measure_prop_name,
            instance_mock(request, MeasureCls, blocks=("A", "B")),
        )
        _assemble_vector_.return_value = np.array([1, 2, 3, 4, 5])
        assembler = StripeAssembler(None, None, None, None)

        value = getattr(assembler, measure_prop_name)

        _assemble_vector_.assert_called_once_with(assembler, ("A", "B"))
        assert value.tolist() == [1, 2, 3, 4, 5]

    def it_knows_the_inserted_row_idxs(self, _row_order_prop_):
        _row_order_prop_.return_value = np.array([-1, 0, 3, -2, 4, 1])
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.inserted_row_idxs == (0, 3)

    def it_knows_the_row_count(self, _row_order_prop_):
        _row_order_prop_.return_value = np.array([1, 2, 3, 4, 5])
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.row_count == 5

    def it_knows_the_row_labels(self, rows_dimension_, _row_order_prop_):
        rows_dimension_.element_labels = ("baz", "foo", "bar")
        rows_dimension_.subtotal_labels = ("bing", "bada")
        _row_order_prop_.return_value = np.array([1, 2, 0, -1, -2])
        assembler = StripeAssembler(None, rows_dimension_, None, None)

        assert assembler.row_labels.tolist() == ["foo", "bar", "baz", "bada", "bing"]

    @pytest.mark.parametrize(
        "order, expected_fills",
        (
            ((2, -2, 0, -1), ("#f00ba5", "STF1", "#000000", "STF2")),
            ((0, 1, 2, -2, -1), ("#000000", "#111111", "#f00ba5", "STF1", "STF2")),
            ((-2, -1, 0, 1, 2), ("STF1", "STF2", "#000000", "#111111", "#f00ba5")),
            ((-1, -2, 2, 1, 0), ("STF2", "STF1", "#f00ba5", "#111111", "#000000")),
        ),
    )
    def it_knows_the_rows_dimension_fills(
        self, request, rows_dimension_, _row_order_prop_, order, expected_fills
    ):
        element_fills = ("#000000", "#111111", "#f00ba5")
        subtotal_fills = ("STF1", "STF2")
        rows_dimension_.valid_elements = tuple(
            instance_mock(request, _Element, fill=fill) for fill in element_fills
        )
        rows_dimension_.subtotals = tuple(
            instance_mock(request, _Subtotal, fill=fill) for fill in subtotal_fills
        )
        _row_order_prop_.return_value = order
        assembler = StripeAssembler(None, rows_dimension_, None, None)

        assert assembler.rows_dimension_fills == expected_fills

    def it_knows_the_scale_mean(self, _measures_prop_, measures_, scaled_counts_):
        scaled_counts_.scale_mean = 3
        measures_.scaled_counts = scaled_counts_
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.scale_mean == 3

    def it_knows_the_scale_median(self, _measures_prop_, measures_, scaled_counts_):
        scaled_counts_.scale_median = 4
        measures_.scaled_counts = scaled_counts_
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.scale_median == 4

    def it_knows_the_scale_stddev(self, _measures_prop_, measures_, scaled_counts_):
        scaled_counts_.scale_stddev = 5
        measures_.scaled_counts = scaled_counts_
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.scale_stddev == 5

    def it_knows_the_scale_stderr(self, _measures_prop_, measures_, scaled_counts_):
        scaled_counts_.scale_stderr = 6
        measures_.scaled_counts = scaled_counts_
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.scale_stderr == 6

    def it_knows_the_table_base_range(self, request, _measures_prop_, measures_):
        measures_.unweighted_bases = instance_mock(
            request, _UnweightedBases, table_base_range=np.array([50, 100])
        )
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.table_base_range.tolist() == [50, 100]

    def it_knows_the_table_margin_range(self, request, _measures_prop_, measures_):
        measures_.weighted_bases = instance_mock(
            request, _WeightedBases, table_margin_range=np.array([50.5, 100.1])
        )
        _measures_prop_.return_value = measures_
        assembler = StripeAssembler(None, None, None, None)

        assert assembler.table_margin_range.tolist() == [50.5, 100.1]

    def it_can_assemble_a_vector_to_help(self, _row_order_prop_):
        base_values = np.array([1, 2, 3, 4])
        subtotal_values = (3, 5, 7)
        blocks = (base_values, subtotal_values)
        _row_order_prop_.return_value = np.array([-3, 1, 0, -2, 3, 2, -1])
        assembler = StripeAssembler(None, None, None, None)

        assert assembler._assemble_vector(blocks).tolist() == [3, 2, 1, 5, 4, 3, 7]

    def it_constructs_its_measures_collaborator_object_to_help(
        self, request, cube_, rows_dimension_, measures_
    ):
        StripeMeasures_ = class_mock(
            request,
            "cr.cube.stripe.assembler.StripeMeasures",
            return_value=measures_,
        )
        assembler = StripeAssembler(
            cube_, rows_dimension_, ca_as_0th=False, slice_idx=7
        )

        measures = assembler._measures

        StripeMeasures_.assert_called_once_with(cube_, rows_dimension_, False, 7)
        assert measures is measures_

    def it_knows_the_row_order_to_help(
        self, request, rows_dimension_, _measures_prop_, measures_
    ):
        _measures_prop_.return_value = measures_
        _BaseOrderHelper_ = class_mock(
            request, "cr.cube.stripe.assembler._BaseOrderHelper"
        )
        _BaseOrderHelper_.display_order.return_value = (-1, 1, -2, 2, -3, 3)
        assembler = StripeAssembler(None, rows_dimension_, None, None)

        row_order = assembler._row_order

        _BaseOrderHelper_.display_order.assert_called_once_with(
            rows_dimension_, measures_
        )
        assert row_order.tolist() == [-1, 1, -2, 2, -3, 3]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _assemble_vector_(self, request):
        return method_mock(request, StripeAssembler, "_assemble_vector")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)

    @pytest.fixture
    def _measures_prop_(self, request):
        return property_mock(request, StripeAssembler, "_measures")

    @pytest.fixture
    def _row_order_prop_(self, request):
        return property_mock(request, StripeAssembler, "_row_order")

    @pytest.fixture
    def rows_dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def scaled_counts_(self, request):
        return instance_mock(request, _ScaledCounts)


class Describe_BaseOrderHelper:
    """Unit-test suite for `cr.cube.stripe.assembler._BaseOrderHelper` object."""

    @pytest.mark.parametrize(
        "collation_method, HelperCls",
        (
            (CM.UNIVARIATE_MEASURE, _SortByMeasureHelper),
            (CM.LABEL, _SortByLabelHelper),
            (CM.EXPLICIT_ORDER, _OrderHelper),
            (CM.PAYLOAD_ORDER, _OrderHelper),
        ),
    )
    def it_dispatches_to_the_right_order_helper(
        self, request, measures_, collation_method, HelperCls
    ):
        rows_dimension_ = instance_mock(
            request,
            Dimension,
            order_spec=instance_mock(
                request, _OrderSpec, collation_method=collation_method
            ),
        )
        order_helper_ = instance_mock(
            request, HelperCls, _display_order=np.array([-2, 1, -1, 2])
        )
        HelperCls_ = class_mock(
            request,
            "cr.cube.stripe.assembler.%s" % HelperCls.__name__,
            return_value=order_helper_,
        )

        display_order = _BaseOrderHelper.display_order(rows_dimension_, measures_)

        HelperCls_.assert_called_once_with(rows_dimension_, measures_)
        assert display_order.tolist() == [-2, 1, -1, 2]

    @pytest.mark.parametrize(
        "pruning_base, expected_value",
        (([1, 1, 1], ()), ([1, 0, 1], (1,)), ([0, 0, 0], (0, 1, 2))),
    )
    def it_knows_the_empty_row_idxs_to_help(
        self, measures_, pruning_base, expected_value
    ):
        measures_.pruning_base = np.array(pruning_base)
        order_helper = _BaseOrderHelper(None, measures_)

        assert order_helper._empty_row_idxs == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measures_(self, request):
        return instance_mock(request, StripeMeasures)


class Describe_OrderHelper:
    """Unit test suite for `cr.cube.stripe.assembler._OrderHelper` object."""

    @pytest.mark.parametrize(
        "collation_method, collator_class_name",
        (
            (CM.PAYLOAD_ORDER, "PayloadOrderCollator"),
            (CM.EXPLICIT_ORDER, "ExplicitOrderCollator"),
        ),
    )
    def it_computes_the_order_of_a_rows_dimension_to_help(
        self, request, collation_method, collator_class_name
    ):
        rows_dimension_ = instance_mock(
            request,
            Dimension,
            order_spec=instance_mock(
                request, _OrderSpec, collation_method=collation_method
            ),
        )
        CollatorCls_ = class_mock(
            request, "cr.cube.stripe.assembler.%s" % collator_class_name
        )
        CollatorCls_.display_order.return_value = (1, -2, 3, 5, -1)
        property_mock(request, _OrderHelper, "_empty_row_idxs", return_value=(2, 4, 6))
        order_helper = _OrderHelper(rows_dimension_, None)

        display_order = order_helper._display_order

        CollatorCls_.display_order.assert_called_once_with(rows_dimension_, (2, 4, 6))
        assert display_order == (1, -2, 3, 5, -1)


class Describe_BaseSortByValueHelper:
    """Unit test suite for `cr.cube.strip.assembler._BaseSortByValueHelper`."""

    def it_computes_the_display_order_to_help(
        self,
        dimension_,
        _element_values_prop_,
        _subtotal_values_prop_,
        _empty_row_idxs_prop_,
        SortByValueCollator_,
    ):
        # --- return type of first two is ndarray in real life, but
        # --- assert_called_once_with() won't match on those, so use list instead.
        _element_values_prop_.return_value = [16, 3, 12]
        _subtotal_values_prop_.return_value = [15, 19]
        _empty_row_idxs_prop_.return_value = ()
        SortByValueCollator_.display_order.return_value = (-1, -2, 0, 2, 1)
        order_helper = _BaseSortByValueHelper(dimension_, None)

        order = order_helper._display_order

        SortByValueCollator_.display_order.assert_called_once_with(
            dimension_, [16, 3, 12], [15, 19], ()
        )
        assert order == (-1, -2, 0, 2, 1)

    def but_it_falls_back_to_payload_order_on_value_error(
        self,
        request,
        dimension_,
        _element_values_prop_,
        _subtotal_values_prop_,
        _empty_row_idxs_prop_,
        SortByValueCollator_,
    ):
        _element_values_prop_.return_value = None
        _subtotal_values_prop_.return_value = None
        _empty_row_idxs_prop_.return_value = (4, 2)
        SortByValueCollator_.display_order.side_effect = ValueError
        PayloadOrderCollator_ = class_mock(
            request, "cr.cube.stripe.assembler.PayloadOrderCollator"
        )
        PayloadOrderCollator_.display_order.return_value = (1, 2, 3, 4)
        order_helper = _BaseSortByValueHelper(dimension_, None)

        order = order_helper._display_order

        PayloadOrderCollator_.display_order.assert_called_once_with(dimension_, (4, 2))
        assert order == (1, 2, 3, 4)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _element_values_prop_(self, request):
        return property_mock(request, _BaseSortByValueHelper, "_element_values")

    @pytest.fixture
    def _empty_row_idxs_prop_(self, request):
        return property_mock(request, _BaseSortByValueHelper, "_empty_row_idxs")

    @pytest.fixture
    def SortByValueCollator_(self, request):
        return class_mock(request, "cr.cube.stripe.assembler.SortByValueCollator")

    @pytest.fixture
    def _subtotal_values_prop_(self, request):
        return property_mock(request, _BaseSortByValueHelper, "_subtotal_values")


class Describe_SortByLabelHelper:
    """Unit test suite for `cr.cube.strip.assembler._SortByLabelHelper`."""

    def it_extracts_the_element_values_to_help(self, dimension_):
        dimension_.element_labels = ["b", "a", "c"]
        order_helper = _SortByLabelHelper(dimension_, None)

        assert order_helper._element_values.tolist() == ["b", "a", "c"]

    def it_extracts_the_subtotal_values_to_help(self, dimension_):
        dimension_.subtotal_labels = ["b", "a", "c"]
        order_helper = _SortByLabelHelper(dimension_, None)

        assert order_helper._subtotal_values.tolist() == ["b", "a", "c"]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_SortByMeasureHelper:
    """Unit test suite for `cr.cube.strip.assembler._SortByMeasureHelper`."""

    def it_extracts_the_element_values_to_help(self, _measure_prop_, measure_):
        _measure_prop_.return_value = measure_
        measure_.blocks = [np.arange(5), None]
        order_helper = _SortByMeasureHelper(None, None)

        assert order_helper._element_values.tolist() == [0, 1, 2, 3, 4]

    @pytest.mark.parametrize(
        "json_name, internal_name",
        (
            ("base_unweighted", "unweighted_bases"),
            ("base_weighted", "weighted_bases"),
            ("count_unweighted", "unweighted_counts"),
            ("count_weighted", "weighted_counts"),
            ("mean", "means"),
            ("percent", "table_proportions"),
            ("percent_stddev", "table_proportion_stddevs"),
            ("percent_moe", "table_proportion_stderrs"),
            ("population", "population_proportions"),
            ("population_moe", "population_proportion_stderrs"),
            ("share_sum", "share_sum"),
            ("sum", "sums"),
        ),
    )
    def it_retrieves_the_right_measure_object_to_help(
        self,
        request,
        _order_spec_prop_,
        order_spec_,
        measure_,
        json_name,
        internal_name,
    ):
        measures_ = instance_mock(request, StripeMeasures)
        setattr(measures_, internal_name, measure_)
        _order_spec_prop_.return_value = order_spec_
        order_spec_.measure_keyname = json_name
        order_helper = _SortByMeasureHelper(None, measures_)

        assert order_helper._measure is measure_

    def but_it_raises_when_the_sort_measure_is_not_supported(
        self, _order_spec_prop_, order_spec_
    ):
        _order_spec_prop_.return_value = order_spec_
        order_spec_.measure_keyname = "foobar"
        order_helper = _SortByMeasureHelper(None, None)

        with pytest.raises(ValueError) as e:
            order_helper._measure

        assert str(e.value) == "sort-by-value for measure 'foobar' is not yet supported"

    def it_extracts_the_subtotal_values_to_help(self, _measure_prop_, measure_):
        _measure_prop_.return_value = measure_
        measure_.blocks = [None, np.arange(3)]
        order_helper = _SortByMeasureHelper(None, None)

        assert order_helper._subtotal_values.tolist() == [0, 1, 2]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def measure_(self, request):
        return instance_mock(request, _BaseSecondOrderMeasure)

    @pytest.fixture
    def _measure_prop_(self, request):
        return property_mock(request, _SortByMeasureHelper, "_measure")

    @pytest.fixture
    def order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, _SortByMeasureHelper, "_order_spec")



# encoding: utf-8

"""Unit test suite for `cr.cube.stripe.cubemeasure` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.stripe.cubemeasure import (
    _BaseCubeMeans,
    _BaseCubeSums,
    _BaseCubeCounts,
    _CatCubeCounts,
    _CatCubeMeans,
    _CatCubeSums,
    CubeMeasures,
    _MrCubeCounts,
    _MrCubeMeans,
    _MrCubeSums,
    _NumArrCubeCounts,
)

from ...unitutil import class_mock, instance_mock, property_mock


class DescribeCubeMeasures:
    """Unit-test suite for `cr.cube.stripe.cubemeasure.CubeMeasures` object."""

    def it_provides_access_to_the_cube_means_object(
        self, request, cube_, rows_dimension_
    ):
        cube_means_ = instance_mock(request, _BaseCubeMeans)
        _BaseCubeMeans_ = class_mock(
            request, "cr.cube.stripe.cubemeasure._BaseCubeMeans"
        )
        _BaseCubeMeans_.factory.return_value = cube_means_
        cube_measures = CubeMeasures(cube_, rows_dimension_, None, None)

        cube_means = cube_measures.cube_means

        _BaseCubeMeans_.factory.assert_called_once_with(cube_, rows_dimension_)
        assert cube_means is cube_means_

    def it_provides_access_to_the_cube_sum_object(
        self, request, cube_, rows_dimension_
    ):
        cube_sum_ = instance_mock(request, _BaseCubeSums)
        _BaseCubeSums_ = class_mock(request, "cr.cube.stripe.cubemeasure._BaseCubeSums")
        _BaseCubeSums_.factory.return_value = cube_sum_
        cube_measures = CubeMeasures(cube_, rows_dimension_, None, None)

        cube_sum = cube_measures.cube_sum

        _BaseCubeSums_.factory.assert_called_once_with(cube_, rows_dimension_)
        assert cube_sum is cube_sum_

    @pytest.mark.parametrize(
        "unweighted_counts, unweighted_valid_counts, counts_used",
        (
            ("counts", None, "counts"),
            ("counts", "valid", "valid"),
        ),
    )
    def it_provides_access_to_the_unweighted_cube_counts_object(
        self,
        cube_,
        cube_counts_,
        _BaseCubeCounts_,
        rows_dimension_,
        unweighted_counts,
        unweighted_valid_counts,
        counts_used,
    ):
        cube_.unweighted_counts = unweighted_counts
        cube_.unweighted_valid_counts = unweighted_valid_counts
        _BaseCubeCounts_.factory.return_value = cube_counts_
        cube_measures = CubeMeasures(cube_, rows_dimension_, False, slice_idx=7)

        unweighted_cube_counts = cube_measures.unweighted_cube_counts

        _BaseCubeCounts_.factory.assert_called_once_with(
            counts_used, rows_dimension_, False, 7
        )
        assert unweighted_cube_counts is cube_counts_

    @pytest.mark.parametrize(
        "weighted_counts, weighted_valid_counts, counts_used",
        (
            ("counts", None, "counts"),
            ("counts", "valid", "valid"),
        ),
    )
    def it_provides_access_to_the_weighted_cube_counts_object(
        self,
        cube_,
        cube_counts_,
        _BaseCubeCounts_,
        rows_dimension_,
        weighted_counts,
        weighted_valid_counts,
        counts_used,
    ):
        cube_.counts = weighted_counts
        cube_.weighted_valid_counts = weighted_valid_counts
        _BaseCubeCounts_.factory.return_value = cube_counts_
        cube_measures = CubeMeasures(cube_, rows_dimension_, False, slice_idx=7)

        weighted_cube_counts = cube_measures.weighted_cube_counts

        _BaseCubeCounts_.factory.assert_called_once_with(
            counts_used, rows_dimension_, False, 7
        )
        assert weighted_cube_counts is cube_counts_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _BaseCubeCounts_(self, request):
        return class_mock(request, "cr.cube.stripe.cubemeasure._BaseCubeCounts")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def cube_counts_(self, request):
        return instance_mock(request, _BaseCubeCounts)

    @pytest.fixture
    def rows_dimension_(self, request):
        return instance_mock(request, Dimension)


# === COUNTS ===


class Describe_BaseCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseUnweightedCubeCounts`."""

    @pytest.mark.parametrize(
        (
            "ca_as_0th",
            "rows_dimension_type",
            "CubeCountsCls",
            "counts",
            "expected_counts",
        ),
        (
            (
                True,
                DT.CA_CAT,
                _CatCubeCounts,
                [[4, 5, 6], [1, 2, 3]],
                [1, 2, 3],
            ),
            (False, DT.MR, _MrCubeCounts, [1, 2, 3], [1, 2, 3]),
            (False, DT.CAT, _CatCubeCounts, [1, 2, 3], [1, 2, 3]),
            (
                False,
                DT.NUM_ARRAY,
                _NumArrCubeCounts,
                [4, 5, 6],
                [4, 5, 6],
            ),
        ),
    )
    def it_provides_a_factory_for_constructing_unweighted_cube_count_objects(
        self,
        request,
        ca_as_0th,
        rows_dimension_type,
        CubeCountsCls,
        counts,
        expected_counts,
    ):
        rows_dimension_ = instance_mock(
            request, Dimension, dimension_type=rows_dimension_type
        )
        cube_counts_ = instance_mock(request, CubeCountsCls)
        CubeCountsCls_ = class_mock(
            request,
            "cr.cube.stripe.cubemeasure.%s" % CubeCountsCls.__name__,
            return_value=cube_counts_,
        )

        unweighted_cube_counts = _BaseCubeCounts.factory(
            counts, rows_dimension_, ca_as_0th, slice_idx=1
        )

        CubeCountsCls_.assert_called_once_with(rows_dimension_, expected_counts)
        assert unweighted_cube_counts is cube_counts_


class Describe_CatCubeCounts:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._CatCubeCounts`."""

    def it_knows_its_bases(self, request, raw_counts):
        property_mock(
            request,
            _CatCubeCounts,
            "table_base",
            return_value=42,
        )
        cube_counts = _CatCubeCounts(None, raw_counts)
        assert cube_counts.bases.tolist() == [42, 42, 42]

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _CatCubeCounts(None, raw_counts)
        assert cube_counts.counts.tolist() == [1, 2, 3]

    def it_knows_its_pruning_base(self, raw_counts):
        cube_counts = _CatCubeCounts(None, raw_counts)
        assert cube_counts.pruning_base.tolist() == [1, 2, 3]

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _CatCubeCounts(None, raw_counts)
        assert cube_counts.table_base == 6

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(3,) np.int ndarray of cube-counts as received from Cube."""
        return np.array([1, 2, 3])


class Describe_MrCubeCounts:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._MrCubeCounts`."""

    def it_knows_its_bases(self, raw_counts):
        cube_counts = _MrCubeCounts(None, raw_counts)
        assert cube_counts.bases.tolist() == [3, 7, 11]

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _MrCubeCounts(None, raw_counts)
        assert cube_counts.counts.tolist() == [1, 3, 5]

    def it_knows_its_pruning_base(self, raw_counts):
        cube_counts = _MrCubeCounts(None, raw_counts)
        assert cube_counts.pruning_base.tolist() == [3, 7, 11]

    def it_knows_its_table_base_is_None(self, raw_counts):
        cube_counts = _MrCubeCounts(None, raw_counts)
        assert cube_counts.table_base is None

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(3, 2) np.int ndarray of unweighted cube-counts as received from Cube."""
        return np.array([[1, 2], [3, 4], [5, 6]])


class Describe_NumArrCubeCounts:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._NumArrCubeCounts`."""

    def it_knows_its_bases(self, raw_counts):
        cube_counts = _NumArrCubeCounts(None, raw_counts)
        assert cube_counts.bases.tolist() == [1, 2, 3]

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _NumArrCubeCounts(None, raw_counts)
        assert cube_counts.counts.tolist() == [1, 2, 3]

    def it_knows_its_pruning_base(self, raw_counts):
        cube_counts = _NumArrCubeCounts(None, raw_counts)
        assert cube_counts.pruning_base.tolist() == [1, 2, 3]

    def it_knows_its_table_base_is_None(self, raw_counts):
        cube_counts = _NumArrCubeCounts(None, raw_counts)
        assert cube_counts.table_base is None

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(3,) np.int ndarray of valid cube-counts as received from Cube."""
        return np.array([1, 2, 3])


# === MEANS ===


class Describe_BaseCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeMeans`."""

    @pytest.mark.parametrize(
        "rows_dimension_type, CubeMeansCls, means",
        (
            (DT.CAT, _CatCubeMeans, [1, 2, 3]),
            (DT.MR, _MrCubeMeans, [[1, 6], [2, 5], [3, 4]]),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_means_objects(
        self, request, rows_dimension_type, CubeMeansCls, means
    ):
        cube_ = instance_mock(request, Cube, means=means)
        rows_dimension_ = instance_mock(
            request, Dimension, dimension_type=rows_dimension_type
        )
        cube_means_ = instance_mock(request, CubeMeansCls)
        CubeMeansCls_ = class_mock(
            request,
            "cr.cube.stripe.cubemeasure.%s" % CubeMeansCls.__name__,
            return_value=cube_means_,
        )

        cube_means = _BaseCubeMeans.factory(cube_, rows_dimension_)

        CubeMeansCls_.assert_called_once_with(rows_dimension_, means)
        assert cube_means is cube_means_


class Describe_CatCubeMeans:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._CatCubeMeans`."""

    def it_knows_its_means(self):
        cube_means = _CatCubeMeans(None, np.array([1.1, 2.2, 3.3]))
        assert cube_means.means == pytest.approx([1.1, 2.2, 3.3])

    def but_it_raises_value_error_when_the_cube_result_does_not_contain_means(
        self, request
    ):
        cube_ = instance_mock(request, Cube)
        cube_.means = None
        with pytest.raises(ValueError) as e:
            _CatCubeMeans(None, None).factory(cube_, None)

        assert str(e.value) == "cube-result does not contain cube-means measure"


class Describe_MrCubeMeans:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._MrCubeMeans`."""

    def it_knows_its_means(self):
        cube_means = _MrCubeMeans(None, np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))
        assert cube_means.means == pytest.approx([1.1, 3.3, 5.5])

    def but_it_raises_value_error_when_the_cube_result_does_not_contain_means(
        self, request
    ):
        cube_ = instance_mock(request, Cube)
        cube_.means = None
        with pytest.raises(ValueError) as e:
            _CatCubeMeans(None, None).factory(cube_, None)

        assert str(e.value) == "cube-result does not contain cube-means measure"


# === SUM ===


class Describe_BaseCubeSums:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeSums`."""

    @pytest.mark.parametrize(
        "rows_dimension_type, CubeSumCls, sums",
        (
            (DT.CAT, _CatCubeSums, [1, 2, 3]),
            (DT.MR, _MrCubeSums, [[1, 6], [2, 5], [3, 4]]),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_sums_objects(
        self, request, rows_dimension_type, CubeSumCls, sums
    ):
        cube_ = instance_mock(request, Cube, sums=sums)
        rows_dimension_ = instance_mock(
            request, Dimension, dimension_type=rows_dimension_type
        )
        cube_sums_ = instance_mock(request, CubeSumCls)
        CubeSumCls_ = class_mock(
            request,
            "cr.cube.stripe.cubemeasure.%s" % CubeSumCls.__name__,
            return_value=cube_sums_,
        )

        cube_sums = _BaseCubeSums.factory(cube_, rows_dimension_)

        CubeSumCls_.assert_called_once_with(rows_dimension_, sums)
        assert cube_sums is cube_sums_


class Describe_CatCubeSums:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._CatCubeSum`."""

    def it_knows_its_sum(self):
        cube_sum = _CatCubeSums(None, np.array([1, 2, 3]))
        assert cube_sum.sums == pytest.approx([1, 2, 3])

    def but_it_raises_value_error_when_the_cube_result_does_not_contain_sums(
        self, request
    ):
        cube_ = instance_mock(request, Cube)
        cube_.sums = None
        with pytest.raises(ValueError) as e:
            _CatCubeSums(None, None).factory(cube_, None)

        assert str(e.value) == "cube-result does not contain cube-sum measure"


class Describe_MrCubeSums:
    """Unit-test suite for `cr.cube.stripe.cubemeasure._MrCubeSum`."""

    def it_knows_its_sum(self):
        cube_sum = _MrCubeSums(None, np.array([[1, 2], [3, 4], [5, 6]]))
        assert cube_sum.sums == pytest.approx([1, 3, 5])

    def but_it_raises_value_error_when_the_cube_result_does_not_contain_sum(
        self, request
    ):
        cube_ = instance_mock(request, Cube)
        cube_.sums = None
        with pytest.raises(ValueError) as e:
            _CatCubeSums(None, None).factory(cube_, None)

        assert str(e.value) == "cube-result does not contain cube-sum measure"



# encoding: utf-8

"""Unit test suite for `cr.cube.matrix.cubemeasure` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _ValidElements
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.matrix.cubemeasure import (
    _ArrXArrCubeCounts,
    _ArrXCatCubeCounts,
    _ArrXMrCubeCounts,
    _BaseCubeCounts,
    _BaseCubeMeans,
    _BaseCubeMeasure,
    _BaseCubeOverlaps,
    _BaseCubeStdDev,
    _BaseCubeSums,
    _BaseUnconditionalCubeCounts,
    _CatXArrCubeCounts,
    _CatXCatCubeCounts,
    _CatXCatCubeMeans,
    _CatXCatCubeStdDev,
    _CatXCatCubeSums,
    _CatXCatUnconditionalCubeCounts,
    _CatXMrCubeCounts,
    _CatXMrCubeMeans,
    _CatXMrCubeStdDev,
    _CatXMrCubeSums,
    _CatXMrUnconditionalCubeCounts,
    CubeMeasures,
    _MrXArrCubeCounts,
    _MrXCatCubeCounts,
    _MrXCatCubeMeans,
    _MrXCatCubeStdDev,
    _MrXCatCubeSums,
    _MrXCatUnconditionalCubeCounts,
    _MrXMrCubeCounts,
    _MrXMrCubeMeans,
    _MrXMrCubeStdDev,
    _MrXMrCubeSums,
    _MrXMrUnconditionalCubeCounts,
)

from ...unitutil import class_mock, instance_mock, method_mock, property_mock


class DescribeCubeMeasures:
    """Unit test suite for `cr.cube.matrix.cubemeasure.CubeMeasures` object."""

    @pytest.mark.parametrize(
        "cube_measure_, CubeMeasureCls",
        (
            ("cube_means", _BaseCubeMeans),
            ("cube_overlaps", _BaseCubeOverlaps),
            ("cube_sum", _BaseCubeSums),
            ("cube_stddev", _BaseCubeStdDev),
        ),
    )
    def it_provides_access_to_cube_measures_object(
        self, request, cube_, dimensions_, cube_measure_, CubeMeasureCls
    ):
        _cube_measure_ = instance_mock(request, CubeMeasureCls)
        CubeMeasureCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeMeasureCls.__name__,
        )
        CubeMeasureCls_.factory.return_value = _cube_measure_
        cube_measures = CubeMeasures(cube_, dimensions_, slice_idx=37)

        cube_measure = getattr(cube_measures, cube_measure_)

        CubeMeasureCls_.factory.assert_called_once_with(cube_, dimensions_, 37)
        assert cube_measure is _cube_measure_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))


class Describe_BaseCubeMeasure:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeMeasure`."""

    @pytest.mark.parametrize(
        "ndim, tab_dim_type, expected_value",
        (
            (2, None, slice(None, None, None)),
            (3, DT.MR, (3, 0)),
            (3, DT.CAT, 3),
        ),
    )
    def it_computes_index_expression_that_selects_measure_values_from_cube_to_help(
        self, request, ndim, tab_dim_type, expected_value
    ):
        cube_ = instance_mock(request, Cube, ndim=ndim, dimension_types=(tab_dim_type,))
        measure = _BaseCubeMeasure(None)

        assert measure._slice_idx_expr(cube_, slice_idx=3) == expected_value


# === COUNTS (UNWEIGHTED & WEIGHTED)


class Describe_BaseCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeCounts`."""

    @pytest.mark.parametrize(
        ("dimension_types", "CubeCountsCls"),
        (
            ((DT.MR, DT.MR), _MrXMrCubeCounts),
            ((DT.MR, DT.CAT), _MrXCatCubeCounts),
            ((DT.CAT, DT.MR), _CatXMrCubeCounts),
            ((DT.CAT, DT.CAT), _CatXCatCubeCounts),
            ((DT.NUM_ARRAY, DT.MR), _ArrXMrCubeCounts),
            ((DT.NUM_ARRAY, DT.CAT), _ArrXCatCubeCounts),
            ((DT.CA_CAT, DT.CA_SUBVAR), _CatXArrCubeCounts),
            ((DT.MR, DT.CA_SUBVAR), _MrXArrCubeCounts),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_count_objects(
        self,
        request,
        cube_,
        dimensions_,
        dimension_types,
        CubeCountsCls,
    ):
        cube_counts_ = instance_mock(request, CubeCountsCls)
        CubeCountsCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeCountsCls.__name__,
            return_value=cube_counts_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeCounts,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types

        cube_counts = _BaseCubeCounts.factory(
            [[1, 2], [3, 4]], "diff_nans", cube_, dimensions_, slice_idx=7
        )

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeCountsCls_.assert_called_once_with(dimensions_, [3, 4], "diff_nans")
        assert cube_counts is cube_counts_

    def it_provides_columns_pruning_mask(self, request):
        property_mock(
            request,
            _BaseCubeCounts,
            "_columns_pruning_base",
            return_value=np.array([0, 2, 0, 4]),
        )
        cube_counts = _BaseCubeCounts(None, None, None)

        assert cube_counts.columns_pruning_mask.tolist() == [True, False, True, False]

    def it_provides_rows_pruning_mask(self, request):
        property_mock(
            request,
            _BaseCubeCounts,
            "_rows_pruning_base",
            return_value=np.array([0, 2, 0, 4]),
        )
        cube_counts = _BaseCubeCounts(None, None, None)

        assert cube_counts.rows_pruning_mask.tolist() == [True, False, True, False]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimensions_(self, request):
        return (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )


class Describe_ArrXArrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._ArrXArrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(raw_counts)

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(raw_counts)

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(raw_counts)

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(raw_counts)

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([4.5, 3.8, 4.9])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([6.7, 6.5]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3) np.float64 ndarray of counts as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_ArrXCatCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._ArrXCatCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(raw_counts)

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(raw_counts)

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array(
                [
                    [6.7, 6.7, 6.7],
                    [6.5, 6.5, 6.5],
                ]
            )
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base == pytest.approx([6.7, 6.5])

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base == pytest.approx([6.7, 6.5])

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array(
                [
                    [6.7, 6.7, 6.7],
                    [6.5, 6.5, 6.5],
                ]
            )
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([4.5, 3.8, 4.9])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([20.1, 19.5]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3) np.float64 ndarray of counts as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_ArrXMrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._ArrXMrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array(
                [
                    [1.1, 2.2, 3.3],
                    [4.4, 5.5, 6.6],
                ]
            )
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(
            np.array(
                [
                    [1.1, 2.2, 3.3],
                    [4.4, 5.5, 6.6],
                ]
            )
        )

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array(
                [
                    [4.8, 6.8, 8.8],
                    [5.9, 7.9, 9.9],
                ]
            )
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array(
                [
                    [4.8, 6.8, 8.8],
                    [5.9, 7.9, 9.9],
                ]
            )
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array(
                # --- sel/not for all rows in a column.
                # --- example: 10.7 == 1.1 + 3.7 + 4.4 + 1.5
                [10.7, 14.7, 18.7]
            )
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _ArrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([20.4, 23.7]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3, 2) np.float ndarray of weighted cube-counts as received from Cube."""
        return np.array(
            [
                # -- sel / not ---------------
                [  # -- row 0 ----------------
                    [1.1, 3.7],  # -- col 0 --
                    [2.2, 4.6],  # -- col 1 --
                    [3.3, 5.5],  # -- col 2 --
                ],
                [  # -- row 1 ----------------
                    [4.4, 1.5],  # -- col 0 --
                    [5.5, 2.4],  # -- col 1 --
                    [6.6, 3.3],  # -- col 2 --
                ],
            ]
        )


class Describe_CatXArrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXArrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array(
                [
                    [4.5, 3.8, 4.9],
                    [4.5, 3.8, 4.9],
                ],
            )
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base == pytest.approx(np.array([4.5, 3.8, 4.9]))

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base == pytest.approx(
            np.array([4.5, 3.8, 4.9])
        )

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(raw_counts)

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(raw_counts)

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array(
                [
                    [4.5, 3.8, 4.9],
                    [4.5, 3.8, 4.9],
                ],
            )
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([9.0, 7.6, 9.8])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([6.7, 6.5]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3) np.float64 ndarray of counts as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXCatCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array(
                [
                    [4.5, 3.8, 4.9],
                    [4.5, 3.8, 4.9],
                ]
            )
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base == pytest.approx(np.array([4.5, 3.8, 4.9]))

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base == pytest.approx(
            np.array([13.2, 13.2, 13.2])
        )

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(raw_counts)

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array(
                [
                    [6.7, 6.7, 6.7],
                    [6.5, 6.5, 6.5],
                ]
            )
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base == pytest.approx(np.array([6.7, 6.5]))

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base == pytest.approx(np.array([13.2, 13.2]))

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base == pytest.approx(np.array([13.2, 13.2]))

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array(
                [
                    [13.2, 13.2, 13.2],
                    [13.2, 13.2, 13.2],
                ]
            )
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([9.0, 7.6, 9.8])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([20.1, 19.5]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3) np.float64 ndarray of counts as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array([[5.5, 7.7, 9.9], [5.5, 7.7, 9.9]])
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base == pytest.approx(np.array([5.5, 7.7, 9.9]))

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base == pytest.approx(
            np.array([10.7, 14.7, 18.7])
        )

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(
            np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
        )

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array([[4.8, 6.8, 8.8], [5.9, 7.9, 9.9]])
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array([[10.7, 14.7, 18.7], [10.7, 14.7, 18.7]])
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array(
                # --- sel/not for all rows in a column.
                # --- example: 10.7 == 1.1 + 3.7 + 4.4 + 1.5
                [10.7, 14.7, 18.7]
            )
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _CatXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(np.array([20.4, 23.7]))

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 2, 3) np.float ndarray of weighted cube-counts as received from Cube."""
        return np.array(
            [
                # -- sel / not ---------------
                [  # -- row 0 ----------------
                    [1.1, 3.7],  # -- col 0 --
                    [2.2, 4.6],  # -- col 1 --
                    [3.3, 5.5],  # -- col 2 --
                ],
                [  # -- row 1 ----------------
                    [4.4, 1.5],  # -- col 0 --
                    [5.5, 2.4],  # -- col 1 --
                    [6.6, 3.3],  # -- col 2 --
                ],
            ]
        )


class Describe_MrXArrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXArrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array([[4.5, 6.5, 8.5], [5.9, 7.9, 9.9]])
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(
            np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
        )

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array([[4.5, 6.5, 8.5], [5.9, 7.9, 9.9]])
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([10.4, 14.4, 18.4])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXArrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(
            # --- sel/not for all columns in a row
            # --- example: 19.5 = 1.1 + 2.2 + 3.3 + 3.4 + 4.3 + 5.2
            np.array([19.5, 23.7])
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 2, 3) np.float ndarray of weighted cube-counts as received from Cube."""
        return np.array(
            [  # --   0    1    2  cols ------
                [  # -- row 0 ----------------
                    [1.1, 2.2, 3.3],  # -- sel
                    [3.4, 4.3, 5.2],  # -- not
                ],
                [  # -- row 1 ----------------
                    [4.4, 5.5, 6.6],  # -- sel
                    [1.5, 2.4, 3.3],  # -- not
                ],
            ]
        )


class Describe_MrXCatCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            np.array([[4.5, 6.5, 8.5], [5.9, 7.9, 9.9]])
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(
            np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
        )

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            np.array([[6.6, 6.6, 6.6], [16.5, 16.5, 16.5]])
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base == pytest.approx(np.array([6.6, 16.5]))

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base == pytest.approx(np.array([19.5, 23.7]))

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            np.array([[19.5, 19.5, 19.5], [23.7, 23.7, 23.7]])
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            np.array([10.4, 14.4, 18.4])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXCatCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(
            # --- sel/not for all columns in a row
            # --- example: 19.5 = 1.1 + 2.2 + 3.3 + 3.4 + 4.3 + 5.2
            np.array([19.5, 23.7])
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 3) np.float64 ndarray of counts as received from Cube."""
        return np.array(
            [  # --   0    1    2  cols ------
                [  # -- row 0 ----------------
                    [1.1, 2.2, 3.3],  # -- sel
                    [3.4, 4.3, 5.2],  # -- not
                ],
                [  # -- row 1 ----------------
                    [4.4, 5.5, 6.6],  # -- sel
                    [1.5, 2.4, 3.3],  # -- not
                ],
            ]
        )


class Describe_MrXMrCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeCounts`."""

    def it_knows_its_column_bases(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.column_bases == pytest.approx(
            # --- sel/not for each column in sel row
            # --- example: 4.6 = 1.1 + 3.5
            np.array([[4.6, 8.0], [4.2, 12.6]])
        )

    def it_knows_its_columns_base(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_base is None

    def it_knows_its_columns_table_base(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.columns_table_base is None

    def it_knows_its_counts(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.counts == pytest.approx(np.array([[1.1, 3.1], [2.8, 7.3]]))

    def it_knows_its_row_bases(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.row_bases == pytest.approx(
            # --- sel/not for each column in sel row
            # --- example: 3.4 = 1.1 + 2.3
            np.array([[3.4, 7.1], [6.3, 15.5]])
        )

    def it_knows_its_rows_base(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_base is None

    def it_knows_its_rows_table_base(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.rows_table_base is None

    def it_knows_its_table_base(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_base is None

    def it_knows_its_table_bases(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts.table_bases == pytest.approx(
            # --- sel/not for all columns in sel/not rows
            # --- example: 12.2 = 1.1+ 2.3 + 3.5+ 5.3
            np.array([[12.2, 18.8], [14.1, 30.6]])
        )

    def it_provides_the_columns_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._columns_pruning_base == pytest.approx(
            # --- sel/not for all rows in a selected column
            # --- example: 8.8 = 1.1 + 3.5 + 2.8 + 1.4
            np.array([8.8, 20.6])
        )

    def it_provides_the_rows_pruning_base_to_help(self, raw_counts):
        cube_counts = _MrXMrCubeCounts(None, raw_counts, None)

        assert cube_counts._rows_pruning_base == pytest.approx(
            # --- sel/not for all columns in a selected row
            # --- example: 10.5 = 1.1 + 2.3 + 3.1 + 4.0
            np.array([10.5, 21.8])
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_counts(self):
        """(2, 2, 2, 2) np.float ndarray of cube-counts as from Cube."""
        return np.array(
            [  # ------ sel/not (col) -----
                [  # -- row 0 -------------
                    [  # -- selected ------
                        [1.1, 2.3],  # -- col 0
                        [3.1, 4.0],  # -- col 1
                    ],
                    [  # -- not selected --
                        [3.5, 5.3],  # -- col 0
                        [4.9, 6.8],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [2.8, 3.5],  # -- col 0
                        [7.3, 8.2],  # -- col 1
                    ],
                    [  # -- not selected --
                        [1.4, 6.4],  # -- col 0
                        [5.3, 9.8],  # -- col 1
                    ],
                ],
            ]
        )


# === MEANS ===


class Describe_BaseCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeMeans`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeMeansCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeMeans),
            ((DT.MR, DT.CAT), _MrXCatCubeMeans),
            ((DT.CAT, DT.MR), _CatXMrCubeMeans),
            ((DT.CAT, DT.CAT), _CatXCatCubeMeans),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_means_objects(
        self, request, dimension_types, CubeMeansCls
    ):
        cube_ = instance_mock(request, Cube)
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_means_ = instance_mock(request, CubeMeansCls)
        CubeMeanCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeMeansCls.__name__,
            return_value=cube_means_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeMeans,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.means = [[1, 2], [3, 4]]

        cube_means = _BaseCubeMeans.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeMeanCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_means is cube_means_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_mean_measure(
        self, cube_
    ):
        cube_.means = None

        with pytest.raises(ValueError) as e:
            _BaseCubeMeans.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-means measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _CatXCatCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3) np.float64 ndarray of means as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _CatXMrCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3, 2) np.float64 ndarray of means as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _MrXCatCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 3, 2) np.int float64 of means as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeMeans:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeMeans`."""

    def it_knows_its_means(self, raw_means):
        cube_means = _MrXMrCubeMeans(None, raw_means)

        assert cube_means.means.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_means(self):
        """(2, 2, 2, 2) np.float64 ndarray of means as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === STD DEV ===


class Describe_BaseCubeStdDev:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeStdDev`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeStdDevCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeStdDev),
            ((DT.MR, DT.CAT), _MrXCatCubeStdDev),
            ((DT.CAT, DT.MR), _CatXMrCubeStdDev),
            ((DT.CAT, DT.CAT), _CatXCatCubeStdDev),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_stddev_objects(
        self, request, dimension_types, CubeStdDevCls, cube_
    ):
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_stddev_ = instance_mock(request, CubeStdDevCls)
        CubeStdDevCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeStdDevCls.__name__,
            return_value=cube_stddev_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeStdDev,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.stddev = [[1, 2], [3, 4]]

        cube_stddev = _BaseCubeStdDev.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeStdDevCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_stddev is cube_stddev_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_stddev_measure(
        self, cube_
    ):
        cube_.stddev = None

        with pytest.raises(ValueError) as e:
            _BaseCubeStdDev.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-stddev measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeStdDev:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _CatXCatCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3) np.float64 ndarray of stddev as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeStdDev:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _CatXMrCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3, 2) np.float64 ndarray of stddev as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeStdDev:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _MrXCatCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 3, 2) np.int float64 of stddev as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeStdDev:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeStdDev`."""

    def it_knows_its_stddev(self, raw_stddev):
        cube_stddev = _MrXMrCubeStdDev(None, raw_stddev)

        assert cube_stddev.stddev.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_stddev(self):
        """(2, 2, 2, 2) np.float64 ndarray of stddev as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === SUM ===


class Describe_BaseCubeSum:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseCubeSum`."""

    @pytest.mark.parametrize(
        "dimension_types, CubeSumCls",
        (
            ((DT.MR, DT.MR), _MrXMrCubeSums),
            ((DT.MR, DT.CAT), _MrXCatCubeSums),
            ((DT.CAT, DT.MR), _CatXMrCubeSums),
            ((DT.CAT, DT.CAT), _CatXCatCubeSums),
        ),
    )
    def it_provides_a_factory_for_constructing_cube_sum_objects(
        self, request, dimension_types, CubeSumCls, cube_
    ):
        dimensions_ = (
            instance_mock(request, Dimension),
            instance_mock(request, Dimension),
        )
        cube_sums_ = instance_mock(request, CubeSumCls)
        CubeSumCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % CubeSumCls.__name__,
            return_value=cube_sums_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseCubeSums,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.sums = [[1, 2], [3, 4]]

        cube_sums = _BaseCubeSums.factory(cube_, dimensions_, slice_idx=7)

        _slice_idx_expr_.assert_called_once_with(cube_, 7)
        CubeSumCls_.assert_called_once_with(dimensions_, [3, 4])
        assert cube_sums is cube_sums_

    def but_it_raises_a_value_error_when_cube_result_does_not_contain_sum_measure(
        self, cube_
    ):
        cube_.sums = None

        with pytest.raises(ValueError) as e:
            _BaseCubeSums.factory(cube_, None, None)

        assert str(e.value) == "cube-result does not contain cube-sum measure"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)


class Describe_CatXCatCubeSums:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXCatCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _CatXCatCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 2.3, 3.3],
            [3.4, 1.5, 1.6],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3) np.float64 ndarray of sums as received from Cube."""
        return np.array(
            [
                [1.1, 2.3, 3.3],
                [3.4, 1.5, 1.6],
            ]
        )


class Describe_CatXMrCubeSum:
    """Unit test suite for `cr.cube.matrix.cubemeasure._CatXMrCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _CatXMrCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 2.2, 3.2],
            [4.3, 5.1, 6.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3, 2) np.float64 ndarray of sums as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXCatCubeSum:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXCatCubeSums`."""

    def it_knows_its_sum(self, raw_sums):
        cube_sum = _MrXCatCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [1.1, 6.1],
            [4.3, 3.1],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 3, 2) np.int float64 of sums as received from Cube."""
        return np.array(
            [  # -- axes are (rows, cols, sel/not) --
                # --sel/not--
                [  # -- row 0 ------------
                    [1.1, 6.1],  # -- col 0 --
                    [2.2, 5.2],  # -- col 1 --
                    [3.2, 4.2],  # -- col 2 --
                ],
                [  # -- row 1 ------------
                    [4.3, 3.1],  # -- col 0 --
                    [5.1, 2.1],  # -- col 1 --
                    [6.1, 1.1],  # -- col 2 --
                    # --------------------
                ],
            ]
        )


class Describe_MrXMrCubeSum:
    """Unit test suite for `cr.cube.matrix.cubemeasure._MrXMrCubeSums`."""

    def it_knows_its_sums(self, raw_sums):
        cube_sum = _MrXMrCubeSums(None, raw_sums)

        assert cube_sum.sums.tolist() == [
            [0.1, 0.1],
            [0.4, 0.5],
        ]

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def raw_sums(self):
        """(2, 2, 2, 2) np.float64 ndarray of sums as from Cube."""
        return np.array(
            # -- axes are (rows, sel/not, cols, sel/not) --
            [
                [  # -- row 0 -------------
                    # --sel/not--
                    [  # -- selected ------
                        [0.1, 0.8],  # -- col 0
                        [0.1, 0.7],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.2, 0.6],  # -- col 0
                        [0.3, 0.5],  # -- col 1
                    ],
                ],
                [  # -- row 1 -------------
                    [  # -- selected ------
                        [0.4, 0.4],  # -- col 0
                        [0.5, 0.3],  # -- col 1
                    ],
                    [  # -- not selected --
                        [0.6, 0.2],  # -- col 0
                        [0.7, 0.1],  # -- col 1
                    ],
                ],
            ]
        )


# === (WEIGHTED) UNCONDITIONAL COUNTS ===


class Describe_BaseUnconditionalCubeCounts:
    """Unit test suite for `cr.cube.matrix.cubemeasure._BaseUnconditionalCubeCounts`."""

    @pytest.mark.parametrize(
        (
            "dimension_types",
            "UnconditinoalCubeCountsCls",
            "counts_with_missings",
            "expected_counts_with_missings",
        ),
        (
            (
                (DT.MR, DT.MR),
                _MrXMrUnconditionalCubeCounts,
                [[1, 2], [3, 4]],
                [3, 4],
            ),
            (
                (DT.MR, DT.CAT),
                _MrXCatUnconditionalCubeCounts,
                [[1, 2], [3, 4]],
                [3, 4],
            ),
            (
                (DT.CAT, DT.MR),
                _CatXMrUnconditionalCubeCounts,
                [[1, 2], [3, 4]],
                [3, 4],
            ),
            (
                (DT.CAT, DT.CAT),
                _CatXCatUnconditionalCubeCounts,
                [[1, 2], [3, 4]],
                [3, 4],
            ),
        ),
    )
    def it_provides_a_factory_for_constructing_weighted_cube_count_objects(
        self,
        request,
        cube_,
        dimensions_,
        dimension_types,
        UnconditinoalCubeCountsCls,
        counts_with_missings,
        expected_counts_with_missings,
    ):
        uncond_cube_counts_ = instance_mock(request, UnconditinoalCubeCountsCls)
        WeightedCubeCountsCls_ = class_mock(
            request,
            "cr.cube.matrix.cubemeasure.%s" % UnconditinoalCubeCountsCls.__name__,
            return_value=uncond_cube_counts_,
        )
        _slice_idx_expr_ = method_mock(
            request,
            _BaseUnconditionalCubeCounts,
            "_slice_idx_expr",
            return_value=1,
            autospec=False,
        )
        cube_.dimension_types = dimension_types
        cube_.counts_with_missings = counts_with_missings

        weighted_cube_counts = _BaseUnconditionalCubeCounts.factory(
            cube_,
            dimensions_,
            slice_idx=2,
        )

        _slice_idx_expr_.assert_called_with(cube_, 2)
        WeightedCubeCountsCls_.assert_called_once_with(
            dimensions_,
            expected_counts_with_missings,
        )
        assert weighted_cube_counts is uncond_cube_counts_

    def it_provides_valid_row_idxs_to_help(self, request, dimensions_):
        valid_elements_ = instance_mock(request, _ValidElements, element_idxs=(0, 2, 4))
        dimensions_[0].valid_elements = valid_elements_
        unconditional_counts = _BaseUnconditionalCubeCounts(dimensions_, None)

        assert unconditional_counts._valid_row_idxs[0].tolist() == [0, 2, 4]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimensions_(self, request):
        return instance_mock(request, Dimension), instance_mock(request, Dimension)
